@@ -1,14 +1,19 @@
 import logging
 import os
-import time
+import pathlib
+import datetime
+import re
 from functools import wraps
-from typing import Dict, Iterable, Optional
+from typing import Dict, Optional, Union, List
 
 import git
 import toml
 from sortedcontainers import SortedDict
 
-from binsync.data import (
+
+from libbs.artifacts import (
+    Artifact,
+    ArtifactFormat,
     Comment,
     Enum,
     Function,
@@ -16,10 +21,13 @@ from binsync.data import (
     GlobalVariable,
     Patch,
     StackVariable,
+    Struct,
 )
-from binsync.data.struct import Struct
+from libbs.artifacts import TomlHexEncoder
+from binsync import __version__ as BS_VERS
 from binsync.core.errors import MetadataNotFoundError
 
+
 l = logging.getLogger(__name__)
 
 
@@ -33,21 +41,33 @@ class ArtifactType:
     ENUM = "enum"
 
 
-def dirty_checker(f):
+#
+# Helper Funcs
+#
+
+def sanitize_name(unsafe_name: str) -> str:
+    """
+    C style name sanitization. Replaces all non-alphanumeric characters with underscores.
+    This should always be used when creating files from named C-like objects in the decompiler, like structs.
+    """
+    return re.sub(r"[^a-zA-Z0-9_]", "_", unsafe_name)
+
+def update_dirty_flag(f):
     @wraps(f)
-    def dirtycheck(self, *args, **kwargs):
+    def _update_dirty_flag(self, *args, **kwargs):
         r = f(self, *args, **kwargs)
         if r is True:
             self._dirty = True
         return r
 
-    return dirtycheck
+    return _update_dirty_flag
 
 
 def update_last_change(f):
     @wraps(f)
     def _update_last_change(self, *args, **kwargs):
         should_set = kwargs.pop('set_last_change', True)
+        from_user = kwargs.pop('from_user', None)
         artifact = args[0]
 
         # make a function if one does not exist
@@ -55,8 +75,12 @@ def update_last_change(f):
             func = self.get_or_make_function(artifact.addr)
 
         if not should_set:
+            from_user_msg = f" from {from_user}" if from_user else ""
+            self.last_commit_msg = f"Merged in {artifact}{from_user_msg}"
             return f(self, *args, **kwargs)
-        artifact.last_change = int(time.time())
+
+        self.last_commit_msg = f"Updated {artifact}"
+        artifact.last_change = datetime.datetime.now(tz=datetime.timezone.utc)
 
         # Comment
         if isinstance(artifact, Comment):
@@ -73,6 +97,10 @@ def update_last_change(f):
             artifact_type = ArtifactType.FUNCTION
             func.last_change = artifact.last_change
 
+        elif isinstance(artifact, Function):
+            artifact_loc = artifact.addr
+            artifact_type = ArtifactType.FUNCTION
+
         # Function Header
         elif isinstance(artifact, FunctionHeader):
             artifact_loc = artifact.addr
@@ -111,29 +139,79 @@ def update_last_change(f):
     return _update_last_change
 
 
+def list_files_in_dir(src: Union[pathlib.Path, git.Tree], dir_name, client=None) -> List[str]:
+    if client and isinstance(src, git.Tree):
+        files = client.list_files_in_tree(src)
+        return [name for name in files if name.startswith(dir_name)]
+
+    # load from filesystem
+    if not src:
+        src = pathlib.Path("")
+
+    if not src.joinpath(pathlib.Path(dir_name)).exists():
+        return []
+
+    dir_name_path = pathlib.Path(dir_name)
+    dir_path = src.joinpath(dir_name)
+    return [
+        str(dir_name_path.joinpath(pathlib.Path(name))) for name in os.listdir(dir_path)
+    ]
+
+
+def file_to_str(src: Union[pathlib.Path, git.Tree], filename, client=None) -> Optional[str]:
+    if client and isinstance(src, git.Tree):
+        file_data = client.load_file_from_tree(src, filename)
+    else:
+        if not src:
+            src = pathlib.Path("")
+
+        src = src.joinpath(filename)
+        if not src.exists():
+            file_data = None
+        else:
+            with open(src, "r") as fp:
+                file_data = fp.read()
+
+    return file_data
+
+
+def toml_file_to_dict(src: Union[pathlib.Path, git.Tree], filename, client=None):
+    file_data = file_to_str(src, filename, client=client)
+    return toml.loads(file_data) if file_data is not None else file_data
+
+
+def toml_file_to_artifact(src: Union[pathlib.Path, git.Tree], filename, artifact_cls, client=None) -> Optional[Artifact]:
+    file_data = file_to_str(src, filename, client=client)
+    return artifact_cls.loads(file_data, fmt=ArtifactFormat.TOML) if file_data is not None else None
+
+
+def toml_file_to_artifacts(src: Union[pathlib.Path, git.Tree], filename, artifact_cls, client=None) -> List[Artifact]:
+    file_data = file_to_str(src, filename, client=client)
+    return artifact_cls.loads_many(file_data, fmt=ArtifactFormat.TOML) if file_data is not None else []
+
+
+#
+# State Defn & Operators
+#
 
 class State:
     """
-    The state.
-
-    :ivar str user:     Name of the user.
-    :ivar int version:  Version of the state, starting from 0.
+    A State object is a collection of artifacts that is tied to a specific user, binary, and time. Normally,
+    this means a state is only tied to one commit in Git.
     """
 
-    def __init__(self, user, version=None, client=None):
+    def __init__(self, user: str, version: str = None, client=None, last_push_time=None, last_commit_msg=None, dirty=True):
         # metadata info
         self.user = user  # type: str
-        self.version = version if version is not None else 0  # type: int
+        self.version = version or str(BS_VERS)
         self.last_push_artifact = -1
         self.last_push_artifact_type = -1
-        self.last_push_time = -1
+        self.last_push_time = last_push_time or datetime.datetime.now(tz=datetime.timezone.utc)
+        self.last_commit_msg = last_commit_msg
 
         # the client
         self.client = client  # type: Optional[Client]
 
-        # state is dirty on creation (metadata)
-        self._dirty = True  # type: bool
-
         # data
         self.functions: Dict[int, Function] = {}
         self.comments: Dict[int, Comment] = {}
@@ -142,6 +220,9 @@ class State:
         self.global_vars: Dict[int, GlobalVariable] = {}
         self.enums: Dict[str, Enum] = {}
 
+        # state is dirty on creation (metadata)
+        self._dirty = dirty  # type: bool
+
     def __eq__(self, other):
         if isinstance(other, State):
             return other.functions == self.functions \
@@ -153,8 +234,7 @@ class State:
         return False
 
     def copy(self):
-        state = State(self.user, version=self.version, client=self.client)
-        state._dirty = False
+        state = State(self.user, version=self.version, client=self.client, last_push_time=self.last_push_time, last_commit_msg=self.last_commit_msg, dirty=self._dirty)
         artifacts = ["functions", "comments", "structs", "patches", "global_vars", "enums"]
         for artifact in artifacts:
             setattr(
@@ -165,181 +245,173 @@ class State:
 
         return state
 
+    def __str__(self):
+        return f"<State: {self.user} " \
+               f"funcs={len(self.functions)} " \
+               f"cmts={len(self.comments)} " \
+               f"globals={len(self.structs) + len(self.global_vars) + len(self.enums)}" \
+               f">"
+
+    def __repr__(self):
+        return self.__str__()
+
     @property
     def dirty(self):
         return self._dirty
 
-    def ensure_dir_exists(self, dir_name):
-        if not os.path.exists(dir_name):
-            os.mkdir(dir_name)
-        if not os.path.isdir(dir_name):
-            raise RuntimeError("Cannot create directory %s. Maybe it conflicts with an existing file?" % dir_name)
+    def _dump_data(self, dst: Union[pathlib.Path, git.IndexFile], filename, data):
+        # dump using Git files
+        if self.client and isinstance(dst, git.IndexFile):
+            self.client.add_data(dst, filename, data)
+            return
 
-    def dump_metadata(self, index):
+        # dump using filesystem
+        if not dst:
+            dst = pathlib.Path("")
+
+        out_path = dst.joinpath(filename)
+        pathlib.Path(out_path).parent.mkdir(parents=True, exist_ok=True)
+        with open(out_path, "wb") as fp:
+            fp.write(data)
+
+    def _delete_data(self, dst: Union[pathlib.Path, git.IndexFile], path):
+        # Delete using Git files
+        if self.client and isinstance(dst, git.IndexFile):
+            dst.remove([str(path)], working_tree=True)
+            return
+
+        # Delete using file system
+        if not dst:
+            path.unlink()
+
+    def dump_metadata(self, dst: Union[pathlib.Path, git.IndexFile]):
         d = {
             "user": self.user,
             "version": self.version,
             "last_push_time": self.last_push_time,
             "last_push_artifact": self.last_push_artifact,
             "last_push_artifact_type": self.last_push_artifact_type,
+            "last_commit_msg": self.last_commit_msg,
         }
-        self.client.add_data(index, 'metadata.toml', toml.dumps(d).encode())
+        self._dump_data(dst, 'metadata.toml', toml.dumps(d, encoder=TomlHexEncoder()).encode())
+
+    def dump(self, dst: Union[pathlib.Path, git.IndexFile]):
+        if isinstance(dst, str):
+            dst = pathlib.Path(dst)
 
-    def dump(self, index: git.IndexFile):
         # dump metadata
-        self.dump_metadata(index)
+        self.dump_metadata(dst)
 
         # dump functions, one file per function in ./functions/
         for addr, func in self.functions.items():
-            path = os.path.join('functions', "%08x.toml" % addr)
-            self.client.add_data(index, path, func.dump().encode())
+            path = pathlib.Path('functions').joinpath("%08x.toml" % addr)
+            self._dump_data(dst, path, func.dumps(fmt=ArtifactFormat.TOML).encode())
+
+        if pathlib.Path(self.client.repo_root + '/functions').exists():
+            for path in pathlib.Path(self.client.repo_root + '/functions').iterdir():
+                file = path.stem
+                address = int(file.split(".")[0], 16)
+                if address not in self.functions.keys():
+                    self._delete_data(dst, path)
 
         # dump structs, one file per struct in ./structs/
         for s_name, struct in self.structs.items():
-            path = os.path.join('structs', f"{s_name}.toml")
-            self.client.add_data(index, path, struct.dump().encode())
+            safe_name = sanitize_name(s_name)
+            path = pathlib.Path('structs').joinpath(f"{safe_name}.toml")
+            self._dump_data(dst, path, struct.dumps(fmt=ArtifactFormat.TOML).encode())
+
+        if pathlib.Path(self.client.repo_root + '/structs').exists():
+            for path in pathlib.Path(self.client.repo_root + '/structs').iterdir():
+                file = path.stem
+                name = file.split(".")[0]
+                if name not in self.structs.keys():
+                    self._delete_data(dst, path)
 
         # dump comments
-        self.client.add_data(index, 'comments.toml', toml.dumps(Comment.dump_many(self.comments)).encode())
+        self._dump_data(dst, 'comments.toml', Comment.dumps_many(list(self.comments.values())).encode())
 
         # dump patches
-        self.client.add_data(index, 'patches.toml', toml.dumps(Patch.dump_many(self.patches)).encode())
+        self._dump_data(dst, 'patches.toml', Patch.dumps_many(list(self.patches.values())).encode())
 
         # dump global vars
-        self.client.add_data(index, 'global_vars.toml', toml.dumps(GlobalVariable.dump_many(self.global_vars)).encode())
+        self._dump_data(dst, 'global_vars.toml', GlobalVariable.dumps_many(list(self.global_vars.values())).encode())
 
         # dump enums
-        self.client.add_data(index, 'enums.toml', toml.dumps(Enum.dump_many(self.enums)).encode())
-
-    @staticmethod
-    def load_toml_from_file(filename, client=None, tree=None):
-        if client:
-            file_data = client.load_file_from_tree(tree, filename)
-        else:
-            with open(filename, "r") as fp:
-                file_data = fp.read()
-
-        return toml.loads(file_data)
+        self._dump_data(dst, 'enums.toml', Enum.dumps_many(list(self.enums.values()), key_attr="name").encode())
 
     @classmethod
-    def parse(cls, tree: git.Tree, version=None, client=None):
-        s = cls(None, client=client)
+    def parse(cls, src: Union[pathlib.Path, git.Tree], client=None):
+        if isinstance(src, str):
+            src = pathlib.Path(src)
+
+        state = cls(None, client=client)
 
         # load metadata
-        try:
-            metadata = cls.load_toml_from_file("metadata.toml", client=client, tree=tree)
-        except:
+        metadata = toml_file_to_dict(src, "metadata.toml", client=client)
+        if metadata is None:
             # metadata is not found
             raise MetadataNotFoundError()
-        s.user = metadata["user"]
-
-        s.version = version if version is not None else metadata["version"]
+        state.user = metadata["user"]
+        state.version = metadata["version"]
+        state.last_push_time = metadata.get("last_push_time", None)
 
         # load functions
-        tree_files = client.list_files_in_tree(tree)
-        function_files = [name for name in tree_files if name.startswith("functions")]
+        function_files = list_files_in_dir(src, "functions", client=client)
         for func_file in function_files:
-            try:
-                func_toml = cls.load_toml_from_file(func_file, client=client, tree=tree)
-            except Exception:
-                pass
-            else:
-                func = Function.load(func_toml)
-                s.functions[func.addr] = func
+            func: Function = toml_file_to_artifact(src, func_file, Function, client=client)
+            state.functions[func.addr] = func
 
         # load comments
-        try:
-            comments_toml = cls.load_toml_from_file("comments.toml", client=client, tree=tree)
-        except:
-            pass
-        else:
-            comments = {}
-            for comment in Comment.load_many(comments_toml):
-                comments[comment.addr] = comment
-            s.comments = comments
+        cmts: List[Comment] = toml_file_to_artifacts(src, "comments.toml", Comment, client=client)
+        state.comments = {cmt.addr: cmt for cmt in cmts}
 
         # load patches
-        try:
-            patches_toml = cls.load_toml_from_file("patches.toml", client=client, tree=tree)
-        except:
-            pass
-        else:
-            patches = {}
-            for patch in Patch.load_many(patches_toml):
-                patches[patch.offset] = patch
-            s.patches = SortedDict(patches)
+        patches: List[Patch] = toml_file_to_artifacts(src, "patches.toml", Patch, client=client)
+        state.patches = {patch.offset: patch for patch in patches}
 
         # load global_vars
-        try:
-            global_vars_toml = cls.load_toml_from_file("global_vars.toml", client=client, tree=tree)
-        except:
-            pass
-        else:
-            global_vars = {}
-            for global_var in GlobalVariable.load_many(global_vars_toml):
-                global_vars[global_var.addr] = global_var
-            s.global_vars = SortedDict(global_vars)
+        gvars: List[GlobalVariable] = toml_file_to_artifacts(src, "global_vars.toml", GlobalVariable, client=client)
+        state.global_vars = {gvar.addr: gvar for gvar in gvars}
 
         # load enums
-        try:
-            enums_toml = cls.load_toml_from_file("enums.toml", client=client, tree=tree)
-        except:
-            pass
-        else:
-            s.enums = {
-                enum.name: enum for enum in Enum.load_many(enums_toml)
-            }
+        enums: List[Enum] = toml_file_to_artifacts(src, "enums.toml", Enum, client=client)
+        state.enums = {enum.name: enum for enum in enums}
 
         # load structs
-        tree_files = client.list_files_in_tree(tree)
-        struct_files = [name for name in tree_files if name.startswith("structs")]
+        struct_files = list_files_in_dir(src, "structs", client=client)
         for struct_file in struct_files:
-            try:
-                struct_toml = cls.load_toml_from_file(struct_file, client=client, tree=tree)
-            except:
-                pass
-            else:
-                struct = Struct.load(struct_toml)
-                s.structs[struct.name] = struct
+            struct: Struct = toml_file_to_artifact(src, struct_file, Struct, client=client)
+            state.structs[struct.name] = struct
 
         # clear the dirty bit
-        s._dirty = False
-
-        return s
-
-    def copy_state(self, target_state=None):
-        if target_state is None:
-            l.warning("Cannot copy an empty state (state == None)")
-            return
-
-        self.functions = target_state.functions.copy()
-        self.comments = target_state.comments.copy()
-        self.patches = target_state.patches.copy()
-        self.structs = target_state.structs.copy()
-        self.global_vars = target_state.global_vars.copy()
-        self.enums = target_state.enums.copy()
-        
-    def save(self):
-        if self.client is None:
-            raise RuntimeError("save(): State.client is None.")
-        self.client.commit_state(self)
+        state._dirty = False
+        return state
 
     #
     # Setters
     #
 
-    @dirty_checker
+    @update_dirty_flag
     @update_last_change
-    def set_function_header(self, func_header: FunctionHeader, set_last_change=True):
-        if self.functions[func_header.addr] == func_header:
+    def set_function(self, function: Function, set_last_change=True, **kwargs):
+        if function.addr in self.functions and self.functions[function.addr] == function:
+            return False
+
+        self.functions[function.addr] = function
+        return True
+
+    @update_dirty_flag
+    @update_last_change
+    def set_function_header(self, func_header: FunctionHeader, set_last_change=True, **kwargs):
+        if func_header.addr in self.functions and self.functions[func_header.addr] == func_header:
             return False
 
         self.functions[func_header.addr].header = func_header
         return True
 
-    @dirty_checker
+    @update_dirty_flag
     @update_last_change
-    def set_comment(self, comment: Comment, set_last_change=True):
+    def set_comment(self, comment: Comment, append=False, set_last_change=True, **kwargs):
         if not comment:
             return False
 
@@ -349,14 +421,19 @@ class State:
             old_cmt = None
 
         if old_cmt != comment:
+            if old_cmt is not None and append:
+                comment.comment = old_cmt.comment + "\n" + comment.comment
+                if set_last_change:
+                    comment.last_change = comment.last_change or old_cmt.last_change
+
             self.comments[comment.addr] = comment
             return True
 
         return False
 
-    @dirty_checker
+    @update_dirty_flag
     @update_last_change
-    def set_patch(self, patch, addr, set_last_change=True):
+    def set_patch(self, patch, addr, set_last_change=True, **kwargs):
         if not patch:
             return False
 
@@ -371,30 +448,30 @@ class State:
 
         return False
 
-    @dirty_checker
+    @update_dirty_flag
     @update_last_change
-    def set_stack_variable(self, variable, offset, func_addr, set_last_change=True):
+    def set_stack_variable(self, variable: StackVariable, set_last_change=True, **kwargs):
         if not variable:
             return False
 
-        func = self.get_function(func_addr)
+        func = self.get_function(variable.addr)
         if not func:
             return False
 
         try:
-            old_var = func.stack_vars[offset]
+            old_var = func.stack_vars[variable.offset]
         except KeyError:
             old_var = None
 
         if old_var != variable:
-            func.stack_vars[offset] = variable
+            func.stack_vars[variable.offset] = variable
             return True
 
         return False
 
-    @dirty_checker
+    @update_dirty_flag
     @update_last_change
-    def set_struct(self, struct: Struct, old_name: Optional[str], set_last_change=True):
+    def set_struct(self, struct: Struct, old_name=None, set_last_change=True, **kwargs):
         """
         Sets a struct in the current state. If old_name is not defined (None), then
         this indicates that the struct has not changed names. In that case, simply overwrite the
@@ -417,17 +494,19 @@ class State:
         if old_name is not None:
             try:
                 del self.structs[old_name]
-                #remove_data(self.client.repo.index, os.path.join('structs', f'{old_name}.toml'))
             except KeyError:
                 pass
 
         # set the new struct
         if struct.name is not None:
             self.structs[struct.name] = struct
+            return True
 
-    @dirty_checker
+        return False
+
+    @update_dirty_flag
     @update_last_change
-    def set_global_var(self, gloabl_var: GlobalVariable, set_last_change=True):
+    def set_global_var(self, gloabl_var: GlobalVariable, set_last_change=True, **kwargs):
         try:
             old_gvar = self.global_vars[gloabl_var.addr]
         except KeyError:
@@ -439,9 +518,9 @@ class State:
 
         return False
 
-    @dirty_checker
+    @update_dirty_flag
     @update_last_change
-    def set_enum(self, enum: Enum, set_last_change=True):
+    def set_enum(self, enum: Enum, set_last_change=True, **kwargs):
         try:
             old_enum = self.enums[enum.name]
         except KeyError:
@@ -467,12 +546,23 @@ class State:
         return func
 
     def get_function(self, addr) -> Function:
-        try:
-            func = self.functions[addr]
-        except KeyError:
-            func = None
+        return self.functions.get(addr, None)
 
-        return func
+    def get_functions(self) -> Dict[int, Function]:
+        return self.functions
+
+    def get_function_header(self, addr) -> Optional[FunctionHeader]:
+        func = self.get_function(addr)
+        if not func:
+            return None
+
+        return func.header
+
+    def get_function_headers(self) -> Dict[int, FunctionHeader]:
+        return {
+            addr: func.header
+            for addr, func in self.functions.items() if func.header
+        }
 
     def get_comment(self, addr) -> Comment:
         try:
@@ -482,17 +572,15 @@ class State:
 
         return cmt
 
-    def get_comments(self) -> Dict[int, Comment]:
-        return self.comments
-
-    def get_func_comments(self, func_addr):
+    def get_func_comments(self, func_addr) -> Dict[int, Comment]:
         try:
             func = self.functions[func_addr]
         except KeyError:
             return {}
 
         return {
-            addr: cmt for addr, cmt in self.comments.items() if addr <= func.addr + func.size
+            addr: cmt for addr, cmt in self.comments.items()
+            if func.addr <= addr <= func.addr + func.size
         }
 
     def get_patch(self, addr) -> Patch:
@@ -533,8 +621,8 @@ class State:
 
         return struct
 
-    def get_structs(self) -> Iterable[Struct]:
-        return self.structs.values()
+    def get_structs(self) -> Dict[str, Struct]:
+        return self.structs
 
     def get_global_var(self, addr):
         try:
@@ -544,6 +632,9 @@ class State:
 
         return gvar
 
+    def get_global_vars(self):
+        return self.global_vars
+
     def get_enum(self, name):
         try:
             enum = self.enums[name]
@@ -553,7 +644,7 @@ class State:
         return enum
 
     def get_enums(self):
-        return self.enums.values()
+        return self.enums
 
     def get_last_push_for_artifact_type(self, artifact_type):
         last_change = -1
@@ -581,49 +672,9 @@ class State:
     # Utils
     #
 
-    def diff_comments(self, other_comments: Dict[int, Comment], diff_range=None):
-        """
-
-        :param other_comments:
-        :param diff_range: [start_addr, end_addr]
-        :return:
-        """
-        diff_dict = {}
-
-        for addr, cmt in self.comments.items():
-            # if addr is less than start or bigger than end
-            if diff_range and (diff_range[0] > addr or addr >= diff_range[1]):
-                continue
-
-            try:
-                other_cmt = other_comments[addr]
-            except KeyError:
-                other_cmt = None
-
-            diff_dict[addr] = cmt.diff(other_cmt)
-
-        for addr, cmt in other_comments.items():
-            if diff_range and (diff_range[0] > addr or addr >= diff_range[1]):
-                continue
-
-            if addr in diff_dict:
-                continue
-
-            diff_dict[addr] = cmt.diff(None)
-
-        return diff_dict
-
     def find_func_for_addr(self, search_addr):
         for func_addr, func in self.functions.items():
             if func.addr <= search_addr < (func.addr + func.size):
                 return func
         else:
             return None
-
-    def find_latest_comment_for_func(self, func: Function) -> Optional[Comment]:
-        cmts = [cmt for addr, cmt in self.comments.items() if addr <= func.addr + func.size]
-        if not cmts:
-            return None
-
-        lastest_cmt = max(cmts, key=lambda c: c.last_change if c.last_change else -1)
-        return lastest_cmt
